package Molior;

use strict;
use warnings;
use YAML;
use File::Spec;
use Math::Trig;
use Molior::Opening;
use 5.010;

our $VERSION = 0.01;
our $PI = atan2 (0,-1);

our $SHARE_DIR = 'share';
$SHARE_DIR = $ENV{SHARE_DIR} if defined $ENV{SHARE_DIR};

=head1 NAME

Molior - Generate 3D building models

=head1 SYNOPSIS

  use Molior;
  use File::DXF;
  use File::DXF::ENTITIES;

  my $molior = new Molior;
  $molior->Read ('extrusion.yml');

  my $dxf = new File::DXF;
  $dxf->{ENTITIES} = new File::DXF::ENTITIES;
  push @{$dxf->{ENTITIES}}, $molior->Extrusion;

  open DXF, '>extrusion.dxf';
  binmode DXF, ":crlf";
  print DXF $dxf->DXF;
  close DXF;

=head1 DESCRIPTION

Takes simplified descriptions of building parts stored in YAML files and
derives 3D geometry in DXF format.

=head1 USAGE

Create a new Molior instance:

  use Molior;
  my $molior = new Molior;

=cut

sub new
{
    my $class = shift;
    $class = ref $class || $class;
    my $self;

    $self = {closed => 0, captop => 1, capbottom => 1, type => '',
             elevation => 0.0, height => 3.0, inner => 0.0, outer => 0.1,
             path => [], openings => [], _db => {}};

    bless $self, $class;
    return $self;
}

=pod

Read a YAML definition from disk:

  $molior->Read ('/path/to/thing.yml') || die;

=cut

sub Read
{
    my $self = shift;
    my $path = shift || return 0;
    my $hash = YAML::LoadFile ($path) || return 0;
    for my $key (keys %{$hash})
    {
        $self->{$key} = $hash->{$key};
    }
    for my $segment (@{$self->{openings}})
    {
        for my $opening (@{$segment})
        {
            bless $opening, 'Molior::Opening';
        }
    }
    return 1;
}

=pod

Write a YAML definition to disk:

  $molior->Write ('/path/to/thing.yml') || die;

=cut

sub Write
{
    my $self = shift;
    my $path = shift || return 0;
    my $hash = {};
    for my $key (keys %{$self})
    {
        next if $key =~ /^_/;
        $hash->{$key} = $self->{$key};
    }
    YAML::DumpFile ($path, $hash) || return 0;
    return 1;
}

=pod

Check if the share directory for part/component database exists:

  my $share_dire = $molior->Share_dir
      || die 'Can't find component folder set SHARE_DIR=/path/to/share';

The path can be set using $ENV{SHARE_DIR} or using the Share_dir() method:

..in bash:

  $ export SHARE_DIR=/usr/local/share/molior

..or in perl:

  $molior->Share_dir ('/usr/share/molior');

=cut

sub Share_dir
{
    my $self = shift;
    $SHARE_DIR = shift if scalar @_;
    return $SHARE_DIR if -d $SHARE_DIR;
    return undef;
}

=pod

Slurp up a list of pre-defined openings from a YAML file, add to $self->{_db} hash

  $molior->Read_Db ('/path/to/openings.yml');

=cut

sub Read_Db
{
    my $self = shift;
    my $path = shift || return 0;
    my @list = YAML::LoadFile ($path);
    for my $item (@list)
    {
        $self->{_db}->{$item->{type}} = $item;
    }
    return 1;
}

# place an opening object into a segment

sub Open
{
    my $self = shift;
    my ($id_segment, $opening) = @_;
    push @{$self->{openings}->[$id_segment]}, $opening;
}

=pod

Shuffle openings and delete the last in a segment if necessary to ensure
there are no overlapping openings or openings beyond the segment.

  $molior->Fix_Segment (2);

Optionally specify a 'border' to set the allowed spacing:

  $molior->Fix_Segment (2, 0.25);

=cut

sub Fix_Segment
{
    my $self = shift;
    my $id_segment = shift;
    my $border = shift || 0.225;

    my $openings = $self->{openings}->[$id_segment] || return 1;
    return 1 unless defined $self->{openings}->[$id_segment]->[0];
    $openings = [sort {$a->{along} <=> $b->{along}} @{$openings}];

    my $length = $self->Length_Segment ($id_segment);

    my $length_openings = 0;
    for (@{$openings}) {$length_openings += $self->{_db}->{$_->{type}}->{width}}
    if ($length_openings + ($border * (scalar(@{$openings})+1)) > $length)
    {
        say 'deleting the last opening and trying again';
        pop @{$self->{openings}->[$id_segment]};
        $self->Fix_Segment ($id_segment, $border);
        return 1;
    }
 
    say "Id: $id_segment, Length: $length, Openings: $length_openings";

    # fix overlaps
    for my $id_opening (0 .. scalar @{$openings} -2)
    {
        my $opening = $openings->[$id_opening];
        my $width = $self->{_db}->{$opening->{type}}->{width};
        my $along_ok = $opening->{along} + $width + $border;
        next unless $along_ok > $openings->[$id_opening +1]->{along};
        say 'fixing overlap';
        $openings->[$id_opening +1]->{along} = $along_ok;
    }

    my $opening_last = $openings->[-1];
    my $width_last = $self->{_db}->{$opening_last->{type}}->{width};

    my $overrun = $opening_last->{along} + $width_last + $border - $length;
    if ($overrun > 0)
    {
        say 'fixing overrun';
        for my $id_opening (0 .. scalar @{$openings} -1)
        {
            $openings->[$id_opening]->{along} -= $overrun;
        }
    }
    my $underrun = $border - $openings->[0]->{along};
    return 1 if $underrun < 0;

    say 'fixing underrun';
    for my $id_opening (reverse (0 .. scalar @{$openings} -2))
    {
        my $opening = $openings->[$id_opening];
        my $width = $self->{_db}->{$opening->{type}}->{width};
        my $along_ok = $openings->[$id_opening +1]->{along} - $border - $width;
        next unless $opening->{along} < $along_ok;
        say 'fixing position';
        my $slide = $along_ok - $opening->{along};
        $slide = $underrun if $slide > $underrun;
        $opening->{along} += $slide;
    }

    return 1;
}

=pod

Query or set the 'type' of this molior instance, e.g. 'molior-wall', or
'molior-extrusion':

  my $type = $molior->Type;
  $molior->Type ('molior-extrusion');

=cut

sub Type
{
    my $self = shift;
    $self->{type} = shift if scalar @_;
    return $self->{type};
}

# Some stuff is only practical with quadrilaterals

sub Is_Quadrilateral
{
    my $self = shift;
    return 0 unless scalar @{$self->{path}} == 4;
    return 0 unless $self->{closed};
    return 1;
}

=pod

Query the number of segments in a path:

  my $int = $molior->Segments;

=cut

sub Segments
{
    my $self = shift;
    my $segments = scalar @{$self->{path}};
    $segments -= 1 unless ($self->{closed});
    return $segments;
}

# get the length of a segment

sub Length_Segment
{
    my $self = shift;
    my $id_segment = shift;
    my $end_vertex = $id_segment +1;
    $end_vertex = 0 if ($end_vertex == scalar @{$self->{path}});
    return _distance_2d ($self->{path}->[$end_vertex], $self->{path}->[$id_segment]);
}

# get the angle of a segment (degrees anticlockwise from y=0)

sub Angle_Segment
{
    my $self = shift;
    my $id_segment = shift;
    my $end_vertex = $id_segment +1;
    $end_vertex = 0 if ($end_vertex == scalar @{$self->{path}});
    return _angle_2d ($self->{path}->[$id_segment], $self->{path}->[$end_vertex]) * 180/$PI;
}

# get the direction of a segment (as a normalised 2d vector)

sub Direction_Segment
{
    my $self = shift;
    my $id_segment = shift;
    my $end_vertex = $id_segment +1;
    $end_vertex = 0 if ($end_vertex == scalar @{$self->{path}});
    my $segment_vector = _subtract_2d ($self->{path}->[$end_vertex],
                                       $self->{path}->[$id_segment]);
    _normalise_2d ($segment_vector);
}

# get the normal vector of a segment (2d pointing out)

sub Normal_Segment
{
    my $self = shift;
    my $id_segment = shift;
    $id_segment = 0 if ($id_segment == scalar @{$self->{path}});
    my $a = $self->Direction_Segment ($id_segment);
    [$a->[1], 0-$a->[0]];
}

sub Corner_coor
{
    my $self = shift;
    my $id = shift;
    $id -= scalar @{$self->{path}} if ($id >= scalar @{$self->{path}});
    $id += scalar @{$self->{path}} if ($id < 0);
    return $self->{path}->[$id];
}

# locate the offset outside corner

sub Corner_out
{
    my $self = shift;
    my $id = shift;

    return $self->Corner_offset ($id, $self->{outer});
}

sub Corner_in
{
    my $self = shift;
    my $id = shift;

    return $self->Corner_offset ($id, 0-$self->{inner});
}

sub Corner_offset
{
    my $self = shift;
    my $id = shift;
    my $distance = shift;

    $id -= scalar @{$self->{path}} if ($id >= scalar @{$self->{path}});

    my $offset_a = _scale_2d ($distance, $self->Normal_Segment ($id -1));
    my $offset_b = _scale_2d ($distance, $self->Normal_Segment ($id));
    
    unless ($self->{closed})
    {
        return _add_2d ($self->{path}->[$id], $offset_a) if $id == (scalar @{$self->{path}} -1);
        return _add_2d ($self->{path}->[$id], $offset_b) if $id == 0;
    }

    return _add_2d ($self->{path}->[$id], $offset_a)
        if abs (_distance_2d ([0,0], _subtract_2d ($offset_a, $offset_b))) < 0.0000000001;

    my $line_a = _points_2line (_add_2d ($self->Corner_coor ($id-1), $offset_a),
                                _add_2d ($self->Corner_coor ($id), $offset_a));

    my $line_b = _points_2line (_add_2d ($self->Corner_coor ($id), $offset_b),
                                _add_2d ($self->Corner_coor ($id+1), $offset_b));

    return _line_intersection ($line_a, $line_b);
}

# get two 3d coordinates locating an opening

sub Opening_coor
{
    my $self = shift;
    my ($id_segment, $id_opening) = @_;

    my $segment_direction = $self->Direction_Segment ($id_segment);
 
    my $along = $self->{openings}->[$id_segment]->[$id_opening]->{along};
    my $type = $self->{openings}->[$id_segment]->[$id_opening]->{type};
    my $up = $self->{openings}->[$id_segment]->[$id_opening]->{up};
    my $width = $self->{_db}->{$type}->{width};
    my $height = $self->{_db}->{$type}->{height};

    my $bottom = $self->{elevation} + $up;
    my $top = $bottom + $height;

    my $a = _add_2d ($self->{path}->[$id_segment], _scale_2d ($along, $segment_direction));
    my $b = _add_2d ($self->{path}->[$id_segment], _scale_2d ($along + $width, $segment_direction));

    return ([@{$a}, $bottom], [@{$b}, $top]);
}

# get two 3d coordinates locating an opening on the outer face

sub Opening_out
{
    my $self = shift;
    my ($id_segment, $id_opening) = @_;
    my ($start, $end) = $self->Opening_coor ($id_segment, $id_opening);

    # get a normal vector perpendicular to the wall surface
    my $normal = $self->Normal_Segment ($id_segment);

    # offset opening to place outer face
    my $outer_offset = _scale_2d ($self->{outer}, $normal);
    my $l_out = [@{_add_2d ([$start->[0], $start->[1]], $outer_offset)}, $start->[2]];
    my $r_out = [@{_add_2d ([$end->[0], $end->[1]], $outer_offset)}, $end->[2]];

    return ($l_out, $r_out);
}

# get two 3d coordinates locating an opening on the inner face

sub Opening_in
{
    my $self = shift;
    my ($id_segment, $id_opening) = @_;
    my ($start, $end) = $self->Opening_coor ($id_segment, $id_opening);

    # get a normal vector perpendicular to the wall surface
    my $normal = $self->Normal_Segment ($id_segment);

    my $inner_offset = _scale_2d (0-$self->{inner}, $normal);
    my $l_in = [@{_add_2d ([$start->[0], $start->[1]], $inner_offset)}, $start->[2]];
    my $r_in = [@{_add_2d ([$end->[0], $end->[1]], $inner_offset)}, $end->[2]];

    return ($l_in, $r_in);
}

=pod

Fetch names of all opening types in use, this is a list of file paths to DXF
files to be inserted as BLOCKS:

  my @path_inserts = $molior->Openings_in_use;

Similarly for directly inserted objects:

  push @path_inserts, $molior->Insert_in_use;

=cut

sub Openings_in_use
{
    my $self = shift;
    my %hash;
    for my $segment (@{$self->{openings}})
    {
        for my $opening (@{$segment})
        {
            $hash{$opening->{type}} = 1;
        }
    }
    return map (File::Spec->catfile ($self->Share_dir, $_ .'.dxf'), keys %hash);
}

sub Insert_in_use
{
    my $self = shift;
    return File::Spec->catfile ($self->Share_dir, $self->{name} .'.dxf');
}

=head2 METHODS THAT GENERATE 3D GEOMETRY

Generate a DXF INSERT fragment for a 'molior-insert' object:

  @dxf = $molior->Insert_insert;

..or DXF 3DFACE entities for a 'molior-floor' object:

  push @dxf, $molior->Floor;

..or a 'molior-ceiling' object:

  push @dxf, $molior->Ceiling;

..or a 'molior-roof' object:

  push @dxf, $molior->Roof;

=cut

sub Insert_insert
{
    my $self = shift;
    my @data;
    push @data, [8, 0];
    push @data, [2, $self->{name}];
    push @data, [10, $self->{location}->[0]];
    push @data, [20, $self->{location}->[1]];
    push @data, [30, $self->{elevation} + $self->{height}];
    push @data, [41, $self->{scale}->[0]] if defined $self->{scale}->[0];
    push @data, [42, $self->{scale}->[1]] if defined $self->{scale}->[0];
    push @data, [43, $self->{scale}->[2]] if defined $self->{scale}->[0];
    push @data, [50, $self->{rotation} || 0];
    return ({type => 'INSERT', _data => [@data]});
}

sub Floor
{
    my $self = shift;
    my @list;
    return 0 unless $self->Is_Quadrilateral;
    my @abcd = map {$self->Corner_in ($_)} (0 .. 3);
    my $bot = $self->{elevation};
    my $top = $self->{elevation} + 0.02;

    return _box (@abcd, $top, $bot);
}

sub Ceiling
{
    my $self = shift;
    my @list;
    return 0 unless $self->Is_Quadrilateral;
    my @abcd = map {$self->Corner_in ($_)} (0 .. 3);
    my $bot = $self->{elevation} + $self->{height} - 0.2;
    my $top = $self->{elevation} + $self->{height};

    return _box (@abcd, $top, $bot);
}

sub Roof
{
    my $self = shift;
    my $gable = $self->{gable};
    my $gradient = $self->{gradient};
    return $self->Duopitch  (0, 0, $gradient) if $gable eq '0000';
    return $self->Monopitch (2, 0, $gradient) if $gable eq '1000';
    return $self->Monopitch (3, 0, $gradient) if $gable eq '0100';
    return $self->Monopitch (0, 0, $gradient) if $gable eq '0010';
    return $self->Monopitch (1, 0, $gradient) if $gable eq '0001';
    return $self->Duopitch  (1, 2, $gradient) if $gable eq '1010';
    return $self->Duopitch  (0, 2, $gradient) if $gable eq '0101';
    return $self->Monopitch (2, 1, $gradient) if $gable eq '1100';
    return $self->Monopitch (3, 1, $gradient) if $gable eq '0110';
    return $self->Monopitch (0, 1, $gradient) if $gable eq '0011';
    return $self->Monopitch (1, 1, $gradient) if $gable eq '1001';
    return $self->Monopitch (0, 2, $gradient) if $gable eq '0111';
    return $self->Monopitch (1, 2, $gradient) if $gable eq '1011';
    return $self->Monopitch (2, 2, $gradient) if $gable eq '1101';
    return $self->Monopitch (3, 2, $gradient) if $gable eq '1110';
    return ();
}

# requires a four sided quadrilateral
# first parameter is number of edge to be the eave (0, 1, 2 or 3)
# second parameter is number of gables
# third parameter is gradient (0.0 is horizontal, 1.0 is 45 degrees, etc...)

sub Monopitch
{
    my $self = shift;
    my $edge = shift || 0;
    my $gables = shift || 0;
    my $gradient = shift || (75/225);
    return 0 unless $self->Is_Quadrilateral;

    my @list;
    my ($a, $b, $c, $d) = map {$self->Corner_out ($_)} (0 + $edge .. 3 + $edge);
    my ($a_in, $b_in, $c_in, $d_in) = map {$self->Corner_in ($_)} (0 + $edge .. 3 + $edge);
    my $bot = $self->{elevation} + $self->{height};

    my $line_front = _points_2line ($a, $b);
    my $line_back = _points_2line ($d, $c);

    my $left = _line_intersection ($line_back, _points_2line ($a, $a_in));
    my $right = _line_intersection ($line_back, _points_2line ($b, $b_in));

    $left = $d if $gables > 0;
    $right = $c if ($gables > 1 || $gables < 0);

    my $ridge_vector = _normalise_2d (_subtract_2d ($left, $right));
    my $ridge_vector_gabled = _normalise_2d (_subtract_2d ($d, $c));

    if (_distance_2d ($ridge_vector, $ridge_vector_gabled) > 1)
    {
        say 'ridge inverts, flipping';
        return $self->Duopitch ($edge+1, -1, $gradient) if $gables == 0;
        return $self->Monopitch ($edge+1, -1, $gradient) if $gables == 1;
        return $self->Monopitch ($edge-1, 1, $gradient) if $gables == -1;
    }

    my $dist_c = _perpendicular_distance ($line_front, $right);
    my $top_c = $bot + ($gradient * $dist_c);
    my $dist_d = _perpendicular_distance ($line_front, $left);
    my $top_d = $bot + ($gradient * $dist_d);

    # top surface
    push @list, _3dface ([@{$a}, $bot],
                         [@{$b}, $bot],
                         [@{$right}, $top_c],
                         [@{$left}, $top_d]);

    # bottom surface
    push @list, _3dface ([@{$d}, $bot],
                         [@{$c}, $bot],
                         [@{$b}, $bot],
                         [@{$a}, $bot]);

    # left wedge
    push @list, _3dface ([@{$left}, $top_d],
                         [@{$d}, $bot],
                         [@{$a}, $bot],
                         [@{$a}, $bot]);

    # right wedge
    push @list, _3dface ([@{$right}, $top_c],
                         [@{$b}, $bot],
                         [@{$b}, $bot],
                         [@{$c}, $bot]);

    # back
    push @list, _3dface ([@{$right}, $top_c],
                         [@{$c}, $bot],
                         [@{$d}, $bot],
                         [@{$left}, $top_d]);

    return @list;
}

sub Duopitch
{
    my $self = shift;
    my $edge = shift || 0;
    my $gables = shift || 0;
    my $gradient = shift || (75/225);
    return 0 unless $self->Is_Quadrilateral;

    my @list;
    my ($a, $b, $c, $d) = map {$self->Corner_out ($_)} (0 + $edge .. 3 + $edge);
    my ($a_in, $b_in, $c_in, $d_in) = map {$self->Corner_in ($_)} (0 + $edge .. 3 + $edge);
    my $bot = $self->{elevation} + $self->{height};

    my $line_front = _points_2line ($a, $b);
    my $line_back = _points_2line ($d, $c);

    my $left = _line_intersection (_points_2line ($a, $a_in), _points_2line ($d, $d_in));
    my $right = _line_intersection (_points_2line ($b, $b_in), _points_2line ($c, $c_in));

    my $line_ridge = _points_2line ($left, $right);

    my $left_gabled = _line_intersection ($line_ridge, _points_2line ($a, $d));
    my $right_gabled = _line_intersection ($line_ridge, _points_2line ($b, $c));

    $left = $left_gabled if $gables > 0;
    $right = $right_gabled if ($gables > 1 || $gables < 0);

    my $ridge_vector = _normalise_2d (_subtract_2d ($left, $right));
    my $ridge_vector_gabled = _normalise_2d (_subtract_2d ($left_gabled, $right_gabled));

    if (_distance_2d ($ridge_vector, $ridge_vector_gabled) > 1)
    {
        say 'ridge inverts, flipping';
        return $self->Duopitch ($edge+1, 0, $gradient) if $gables == 0;
        return $self->Monopitch ($edge+1, 0, $gradient) if $gables == 1;
        return $self->Monopitch ($edge-1, 0, $gradient) if $gables == -1;
    }

    my $dist_left = _perpendicular_distance ($line_front, $left);
    my $top_left = $bot + ($gradient * $dist_left);
    my $dist_right = _perpendicular_distance ($line_front, $right);
    my $top_right = $bot + ($gradient * $dist_right);

    # top front surface
    push @list, _3dface ([@{$a}, $bot],
                         [@{$b}, $bot],
                         [@{$right}, $top_right],
                         [@{$left}, $top_left]);
    # top back surface
    push @list, _3dface ([@{$c}, $bot],
                         [@{$d}, $bot],
                         [@{$left}, $top_left],
                         [@{$right}, $top_right]);
    # bottom surface
    push @list, _3dface ([@{$d}, $bot],
                         [@{$c}, $bot],
                         [@{$b}, $bot],
                         [@{$a}, $bot]);
    # left wedge
    push @list, _3dface ([@{$d}, $bot],
                         [@{$a}, $bot],
                         [@{$left}, $top_left],
                         [@{$left}, $top_left]);
    # right wedge
    push @list, _3dface ([@{$b}, $bot],
                         [@{$c}, $bot],
                         [@{$right}, $top_right],
                         [@{$right}, $top_right]);

    return @list;
}

=pod

Generate DXF INSERT and 3DFACE fragments for an opening in a wall, i.e. the window and the wall above and below the window:

  push @dxf, $molior->Inserts;

=cut

sub Inserts
{
    my $self = shift;
    my @list;
    for my $id_segment (0 .. scalar @{$self->{openings}} -1)
    {
        next unless defined $self->{openings}->[$id_segment];
        for my $id_opening (0 .. scalar @{$self->{openings}->[$id_segment]} -1)
        {
            my $opening = $self->{openings}->[$id_segment]->[$id_opening];
            my ($start, $end) = $self->Opening_coor ($id_segment, $id_opening);

            # drop in a block reference for the opening
            my @data;
            push @data, [8, 0];
            push @data, [2, $opening->{type}];
            push @data, [10, $start->[0]];
            push @data, [20, $start->[1]];
            push @data, [30, $start->[2]];
            push @data, [50, $self->Angle_Segment ($id_segment)];
            push @list, {type => 'INSERT', _data => [@data]};

            my ($l_out, $r_out) = $self->Opening_out ($id_segment, $id_opening);
            my ($l_in, $r_in) = $self->Opening_in ($id_segment, $id_opening);

            my $top = $self->{elevation} + $self->{height};
            my $bot = $self->{elevation};

            # outer vertical faces
            push @list, _vertical ([$l_out->[0], $l_out->[1], $bot],
                                   [$r_out->[0], $r_out->[1], $l_out->[2]]);

            push @list, _vertical ([$l_out->[0], $l_out->[1], $r_out->[2]],
                                   [$r_out->[0], $r_out->[1], $top]);

            # inner vertical faces
            push @list, _vertical ([$r_in->[0], $r_in->[1], $bot],
                                   [$l_in->[0], $l_in->[1], $l_in->[2]]);

            push @list, _vertical ([$r_in->[0], $r_in->[1], $r_in->[2]],
                                   [$l_in->[0], $l_in->[1], $top]);

            # top face
            push @list, _3dface ([$l_out->[0], $l_out->[1], $top],
                                 [$r_out->[0], $r_out->[1], $top],
                                 [$r_in->[0], $r_in->[1], $top],
                                 [$l_in->[0], $l_in->[1], $top]) if $self->{captop};

            # bottom face
            push @list, _3dface ([$r_out->[0], $r_out->[1], $bot],
                                 [$l_out->[0], $l_out->[1], $bot],
                                 [$l_in->[0], $l_in->[1], $bot],
                                 [$r_in->[0], $r_in->[1], $bot]) if $self->{capbottom};
        }
    }
    return @list;
}

=pod

Generate DXF 3DFACE fragments for walls, a single gap is left for one or more
openings in each segment:

  push @dxf, $molior->Walls;

=cut

sub Walls
{
    my $self = shift;
    my @list;
    my $top = $self->{elevation} + $self->{height};
    my $bot = $self->{elevation};
    my $last_segment = scalar @{$self->{path}} -2;
    $last_segment += 1 if ($self->{closed});

    # cap the ends of the path

    unless ($self->{closed})
    {
        my $v_out_a = $self->Corner_out (0);
        my $v_in_a = $self->Corner_in (0);

        $v_out_a = _add_2d ($v_out_a, $self->Extension_start);
        $v_in_a = _add_2d ($v_in_a, $self->Extension_start);

        push @list, _vertical ([$v_in_a->[0], $v_in_a->[1], $bot],
                           [$v_out_a->[0], $v_out_a->[1], $top]);

        my $v_out_b = $self->Corner_out (scalar @{$self->{path}} -1);
        my $v_in_b = $self->Corner_in (scalar @{$self->{path}} -1);

        $v_out_b = _add_2d ($v_out_b, $self->Extension_end);
        $v_in_b = _add_2d ($v_in_b, $self->Extension_end);

        push @list, _vertical ([$v_out_b->[0], $v_out_b->[1], $bot],
                               [$v_in_b->[0], $v_in_b->[1], $top]);
    }

    # do segments without any openings
    for my $id_segment (0 .. $last_segment)
    {
        # draw a single face if there are no openings
        next if defined $self->{openings}->[$id_segment]->[0];

        my $v_out_a = $self->Corner_out ($id_segment);
        my $v_out_b = $self->Corner_out ($id_segment +1);

        $v_out_a = _add_2d ($v_out_a, $self->Extension_start)
            if ($id_segment == 0 and !$self->{closed});
        $v_out_b = _add_2d ($v_out_b, $self->Extension_end)
            if ($id_segment == $last_segment and !$self->{closed});

        push @list, _vertical ([$v_out_a->[0], $v_out_a->[1], $bot],
                               [$v_out_b->[0], $v_out_b->[1], $top]);

        my $v_in_a = $self->Corner_in ($id_segment);
        my $v_in_b = $self->Corner_in ($id_segment +1);

        $v_in_a = _add_2d ($v_in_a, $self->Extension_start)
            if ($id_segment == 0 and !$self->{closed});
        $v_in_b = _add_2d ($v_in_b, $self->Extension_end)
            if ($id_segment == $last_segment and !$self->{closed});

        push @list, _vertical ([$v_in_b->[0], $v_in_b->[1], $bot],
                               [$v_in_a->[0], $v_in_a->[1], $top]);

        push @list, _3dface ([$v_in_a->[0], $v_in_a->[1], $top],
                             [$v_out_a->[0], $v_out_a->[1], $top],
                             [$v_out_b->[0], $v_out_b->[1], $top],
                             [$v_in_b->[0], $v_in_b->[1], $top]) if $self->{captop};

        push @list, _3dface ([$v_in_b->[0], $v_in_b->[1], $bot],
                             [$v_out_b->[0], $v_out_b->[1], $bot],
                             [$v_out_a->[0], $v_out_a->[1], $bot],
                             [$v_in_a->[0], $v_in_a->[1], $bot]) if $self->{capbottom};
    }

    for my $id_segment (0 .. $last_segment)
    {
        # draw six faces on each side if there are openings
        next unless defined $self->{openings}->[$id_segment]->[0];

        # establish coordinates for outer faces
        my $v_out_a = $self->Corner_out ($id_segment);
        my $v_out_b = $self->Corner_out ($id_segment +1);

        $v_out_a = _add_2d ($v_out_a, $self->Extension_start)
            if ($id_segment == 0 and !$self->{closed});
        $v_out_b = _add_2d ($v_out_b, $self->Extension_end)
            if ($id_segment == $last_segment and !$self->{closed});

        my ($l_out_a, $r_out_a) = $self->Opening_out ($id_segment, 0);
        my ($l_out_b, $r_out_b) = $self->Opening_out ($id_segment, -1);

        push @list, _3dface ([$l_out_a->[0], $l_out_a->[1], $bot],
                             [$l_out_a->[0], $l_out_a->[1], $l_out_a->[2]],
                             [$v_out_a->[0], $v_out_a->[1], $bot],
                             [$v_out_a->[0], $v_out_a->[1], $bot]);

        push @list, _3dface ([$l_out_a->[0], $l_out_a->[1], $l_out_a->[2]],
                             [$l_out_a->[0], $l_out_a->[1], $r_out_a->[2]],
                             [$v_out_a->[0], $v_out_a->[1], $top],
                             [$v_out_a->[0], $v_out_a->[1], $bot]);

        push @list, _3dface ([$l_out_a->[0], $l_out_a->[1], $r_out_a->[2]],
                             [$l_out_a->[0], $l_out_a->[1], $top],
                             [$v_out_a->[0], $v_out_a->[1], $top],
                             [$v_out_a->[0], $v_out_a->[1], $top]);

        push @list, _3dface ([$r_out_b->[0], $r_out_b->[1], $l_out_b->[2]],
                             [$r_out_b->[0], $r_out_b->[1], $bot],
                             [$v_out_b->[0], $v_out_b->[1], $bot],
                             [$v_out_b->[0], $v_out_b->[1], $bot]);

        push @list, _3dface ([$v_out_b->[0], $v_out_b->[1], $bot],
                             [$v_out_b->[0], $v_out_b->[1], $top],
                             [$r_out_b->[0], $r_out_b->[1], $r_out_b->[2]],
                             [$r_out_b->[0], $r_out_b->[1], $l_out_b->[2]]);

        push @list, _3dface ([$v_out_b->[0], $v_out_b->[1], $top],
                             [$v_out_b->[0], $v_out_b->[1], $top],
                             [$r_out_b->[0], $r_out_b->[1], $top],
                             [$r_out_b->[0], $r_out_b->[1], $r_out_b->[2]]);

        # establish coordinates for inner faces
        my $v_in_a = $self->Corner_in ($id_segment);
        my $v_in_b = $self->Corner_in ($id_segment +1);

        $v_in_a = _add_2d ($v_in_a, $self->Extension_start)
            if ($id_segment == 0 and !$self->{closed});
        $v_in_b = _add_2d ($v_in_b, $self->Extension_end)
            if ($id_segment == $last_segment and !$self->{closed});

        my ($l_in_a, $r_in_a) = $self->Opening_in ($id_segment, 0);
        my ($l_in_b, $r_in_b) = $self->Opening_in ($id_segment, -1);

        push @list, _3dface ([$v_in_a->[0], $v_in_a->[1], $bot],
                             [$l_in_a->[0], $l_in_a->[1], $l_in_a->[2]],
                             [$l_in_a->[0], $l_in_a->[1], $bot],
                             [$l_in_a->[0], $l_in_a->[1], $bot]);

        push @list, _3dface ([$v_in_a->[0], $v_in_a->[1], $bot],
                             [$v_in_a->[0], $v_in_a->[1], $top],
                             [$l_in_a->[0], $l_in_a->[1], $r_in_a->[2]],
                             [$l_in_a->[0], $l_in_a->[1], $l_in_a->[2]]);

        push @list, _3dface ([$v_in_a->[0], $v_in_a->[1], $top],
                             [$v_in_a->[0], $v_in_a->[1], $top],
                             [$l_in_a->[0], $l_in_a->[1], $top],
                             [$l_in_a->[0], $l_in_a->[1], $r_in_a->[2]]);

        push @list, _3dface ([$v_in_b->[0], $v_in_b->[1], $bot],
                             [$r_in_b->[0], $r_in_b->[1], $bot],
                             [$r_in_b->[0], $r_in_b->[1], $l_in_b->[2]],
                             [$r_in_b->[0], $r_in_b->[1], $l_in_b->[2]]);

        push @list, _3dface ([$r_in_b->[0], $r_in_b->[1], $l_in_b->[2]],
                             [$r_in_b->[0], $r_in_b->[1], $r_in_b->[2]],
                             [$v_in_b->[0], $v_in_b->[1], $top],
                             [$v_in_b->[0], $v_in_b->[1], $bot]);

        push @list, _3dface ([$r_in_b->[0], $r_in_b->[1], $r_in_b->[2]],
                             [$r_in_b->[0], $r_in_b->[1], $top],
                             [$v_in_b->[0], $v_in_b->[1], $top],
                             [$v_in_b->[0], $v_in_b->[1], $top]);

        # draw top and bottom faces
        push @list, _3dface ([$l_out_a->[0], $l_out_a->[1], $top],
                             [$l_in_a->[0], $l_in_a->[1], $top],
                             [$v_in_a->[0], $v_in_a->[1], $top],
                             [$v_out_a->[0], $v_out_a->[1], $top]) if $self->{captop};

        push @list, _3dface ([$v_out_a->[0], $v_out_a->[1], $bot],
                             [$v_in_a->[0], $v_in_a->[1], $bot],
                             [$l_in_a->[0], $l_in_a->[1], $bot],
                             [$l_out_a->[0], $l_out_a->[1], $bot]) if $self->{capbottom};

        push @list, _3dface ([$r_out_b->[0], $r_out_b->[1], $top],
                             [$v_out_b->[0], $v_out_b->[1], $top],
                             [$v_in_b->[0], $v_in_b->[1], $top],
                             [$r_in_b->[0], $r_in_b->[1], $top]) if $self->{captop};

        push @list, _3dface ([$r_in_b->[0], $r_in_b->[1], $bot],
                             [$v_in_b->[0], $v_in_b->[1], $bot],
                             [$v_out_b->[0], $v_out_b->[1], $bot],
                             [$r_out_b->[0], $r_out_b->[1], $bot]) if $self->{capbottom};
    }
    return @list;
}

=pod

Generate DXF 3DFACE fragments for the bits of walls between openings:

  push @dxf, $molior->Infills;

=cut

sub Infills
{
    my $self = shift;
    my @list;
    for my $id_segment (0 .. scalar @{$self->{openings}} -1)
    {
        next unless defined $self->{openings}->[$id_segment];
        for my $id_opening (0 .. scalar @{$self->{openings}->[$id_segment]} -2)
        {
            # two openings (a & b)
            # each has inner and outer offsets (in & out)
            # each has bottom-left and top-right 3d points (l & r) 

            my ($l_out_a, $r_out_a) = $self->Opening_out ($id_segment, $id_opening);
            my ($l_in_a, $r_in_a) = $self->Opening_in ($id_segment, $id_opening);

            my ($l_out_b, $r_out_b) = $self->Opening_out ($id_segment, $id_opening +1);
            my ($l_in_b, $r_in_b) = $self->Opening_in ($id_segment, $id_opening +1);

            # top and bot define the top and bottom elevation of the wall itself

            my $top = $self->{elevation} + $self->{height};
            my $bot = $self->{elevation};

            # outer vertical faces
            push @list, _3dface ([$r_out_a->[0], $r_out_a->[1], $bot],
                                 [$l_out_b->[0], $l_out_b->[1], $bot],
                                 [$l_out_b->[0], $l_out_b->[1], $l_out_b->[2]],
                                 [$r_out_a->[0], $r_out_a->[1], $l_out_a->[2]]);

            push @list, _3dface ([$r_out_a->[0], $r_out_a->[1], $l_out_a->[2]],
                                 [$l_out_b->[0], $l_out_b->[1], $l_out_b->[2]],
                                 [$l_out_b->[0], $l_out_b->[1], $r_out_b->[2]],
                                 [$r_out_a->[0], $r_out_a->[1], $r_out_a->[2]]);

            push @list, _3dface ([$r_out_a->[0], $r_out_a->[1], $r_out_a->[2]],
                                 [$l_out_b->[0], $l_out_b->[1], $r_out_b->[2]],
                                 [$l_out_b->[0], $l_out_b->[1], $top],
                                 [$r_out_a->[0], $r_out_a->[1], $top]);

            # inner vertical faces
            push @list, _3dface ([$l_in_b->[0], $l_in_b->[1], $bot],
                                 [$r_in_a->[0], $r_in_a->[1], $bot],
                                 [$r_in_a->[0], $r_in_a->[1], $l_in_a->[2]],
                                 [$l_in_b->[0], $l_in_b->[1], $l_in_b->[2]]);

            push @list, _3dface ([$l_in_b->[0], $l_in_b->[1], $l_in_b->[2]],
                                 [$r_in_a->[0], $r_in_a->[1], $l_in_a->[2]],
                                 [$r_in_a->[0], $r_in_a->[1], $r_in_a->[2]],
                                 [$l_in_b->[0], $l_in_b->[1], $r_in_b->[2]]);

            push @list, _3dface ([$l_in_b->[0], $l_in_b->[1], $r_in_b->[2]],
                                 [$r_in_a->[0], $r_in_a->[1], $r_in_a->[2]],
                                 [$r_in_a->[0], $r_in_a->[1], $top],
                                 [$l_in_b->[0], $l_in_b->[1], $top]);

            # top face
            push @list, _3dface ([$r_out_a->[0], $r_out_a->[1], $top],
                                 [$l_out_b->[0], $l_out_b->[1], $top],
                                 [$l_in_b->[0], $l_in_b->[1], $top],
                                 [$r_in_a->[0], $r_in_a->[1], $top]) if $self->{captop};

            # bottom face
            push @list, _3dface ([$r_out_a->[0], $r_out_a->[1], $bot],
                                 [$r_in_a->[0], $r_in_a->[1], $bot],
                                 [$l_in_b->[0], $l_in_b->[1], $bot],
                                 [$l_out_b->[0], $l_out_b->[1], $bot]) if $self->{capbottom};
        }
    }
    return @list;
}

sub Extension_start
{
    my $self = shift;
    return [0,0] unless defined $self->{extension};
    return _scale_2d ($self->Direction_Segment (0), 0 - $self->{extension});
}

sub Extension_end
{
    my $self = shift;
    return [0,0] unless defined $self->{extension};
    return _scale_2d ($self->Direction_Segment (scalar @{$self->{path}} -2), $self->{extension});
}

=pod

Generate DXF 3DFACE fragments for an extrusion defined in a 'molior-extrusion' object:

  push @dxf, $molior->Extrusion;

=cut

sub Extrusion
{
    my $self = shift;
    my @list;

    my $elevation = $self->{elevation};
    my $height = $self->{height};

    my $last_segment = scalar @{$self->{path}} -2;
    $last_segment += 1 if ($self->{closed});

    # cap the ends of the path
    for my $face (@{$self->{cap}})
    {
        if (!$self->{closed} and scalar @{$face} == 4)
        {
            my $p_a = $self->{profile}->[$face->[0]];
            my $p_b = $self->{profile}->[$face->[1]];
            my $p_c = $self->{profile}->[$face->[2]];
            my $p_d = $self->{profile}->[$face->[3]];

            my $a = $self->Corner_offset (0, $p_a->[0]);
            my $b = $self->Corner_offset (0, $p_b->[0]);
            my $c = $self->Corner_offset (0, $p_c->[0]);
            my $d = $self->Corner_offset (0, $p_d->[0]);

            $a = _add_2d ($a, $self->Extension_start);
            $b = _add_2d ($b, $self->Extension_start);
            $c = _add_2d ($c, $self->Extension_start);
            $d = _add_2d ($d, $self->Extension_start);

            push @list, _3dface ([@{$a}, $p_a->[1] + $elevation + $height],
                                 [@{$b}, $p_b->[1] + $elevation + $height],
                                 [@{$c}, $p_c->[1] + $elevation + $height],
                                 [@{$d}, $p_d->[1] + $elevation + $height]);

            $a = $self->Corner_offset ($last_segment +1, $p_a->[0]);
            $b = $self->Corner_offset ($last_segment +1, $p_b->[0]);
            $c = $self->Corner_offset ($last_segment +1, $p_c->[0]);
            $d = $self->Corner_offset ($last_segment +1, $p_d->[0]);

            $a = _add_2d ($a, $self->Extension_end);
            $b = _add_2d ($b, $self->Extension_end);
            $c = _add_2d ($c, $self->Extension_end);
            $d = _add_2d ($d, $self->Extension_end);

            push @list, _3dface ([@{$d}, $p_d->[1] + $elevation + $height],
                                 [@{$c}, $p_c->[1] + $elevation + $height],
                                 [@{$b}, $p_b->[1] + $elevation + $height],
                                 [@{$a}, $p_a->[1] + $elevation + $height]);
        }
    }

    for my $id_segment (0 .. $last_segment)
    {
        for my $id_point ( -1 .. scalar @{$self->{profile}} -2)
        {
            my $m = $self->{profile}->[$id_point];
            my $n = $self->{profile}->[$id_point +1];

            my $a = $self->Corner_offset ($id_segment, $n->[0]);
            my $b = $self->Corner_offset ($id_segment, $m->[0]);
            my $c = $self->Corner_offset ($id_segment +1, $m->[0]);
            my $d = $self->Corner_offset ($id_segment +1, $n->[0]);

            if (!$self->{closed} and $id_segment == 0)
            {
                $a = _add_2d ($a, $self->Extension_start);
                $b = _add_2d ($b, $self->Extension_start);
            }
            if (!$self->{closed} and $id_segment == $last_segment)
            {
                $c = _add_2d ($c, $self->Extension_end);
                $d = _add_2d ($d, $self->Extension_end);
            }

            push @list, _3dface ([@{$a}, $n->[1] + $elevation + $height],
                                 [@{$b}, $m->[1] + $elevation + $height],
                                 [@{$c}, $m->[1] + $elevation + $height],
                                 [@{$d}, $n->[1] + $elevation + $height]);
        }
    }

    return @list;
}

# UTILITY FUNCTIONS

# draw a quad face
# generally interpreted as two triangles: a,b,d and b,c,d

sub _3dface
{
    my ($a, $b, $c, $d) = @_;
    my @data;
    push @data, [8, 0];

    push @data, [10, $a->[0]];
    push @data, [20, $a->[1]];
    push @data, [30, $a->[2]];

    push @data, [11, $b->[0]];
    push @data, [21, $b->[1]];
    push @data, [31, $b->[2]];

    push @data, [12, $c->[0]];
    push @data, [22, $c->[1]];
    push @data, [32, $c->[2]];

    push @data, [13, $d->[0]];
    push @data, [23, $d->[1]];
    push @data, [33, $d->[2]];
    return {type => '3DFACE', _data => [@data]};
}

# draw a vertical quad face

sub _vertical
{
    my ($a, $b) = @_;
    my @data;
    push @data, [8, 0];

    push @data, [10, $a->[0]];
    push @data, [20, $a->[1]];
    push @data, [30, $a->[2]];

    push @data, [11, $b->[0]];
    push @data, [21, $b->[1]];
    push @data, [31, $a->[2]];

    push @data, [12, $b->[0]];
    push @data, [22, $b->[1]];
    push @data, [32, $b->[2]];

    push @data, [13, $a->[0]];
    push @data, [23, $a->[1]];
    push @data, [33, $b->[2]];
    return {type => '3DFACE', _data => [@data]};
}

# draw a box that is quadrilateral in plan with vertical sides and horizontal top/bottom

sub _box
{
    my ($a, $b, $c, $d, $top, $bot) = @_;
    my @list;
     
    # top face
    push @list, _3dface ([@{$a}, $top],
                         [@{$b}, $top],
                         [@{$c}, $top],
                         [@{$d}, $top]);

    # bottom face
    push @list, _3dface ([@{$d}, $bot],
                         [@{$c}, $bot],
                         [@{$b}, $bot],
                         [@{$a}, $bot]);

    # outer faces
    push @list, _vertical ([@{$a}, $bot],
                           [@{$b}, $top]);
    push @list, _vertical ([@{$b}, $bot],
                           [@{$c}, $top]);
    push @list, _vertical ([@{$c}, $bot],
                           [@{$d}, $top]);
    push @list, _vertical ([@{$d}, $bot],
                           [@{$a}, $top]);

    return @list;
}

# vector math

sub _distance_2d
{
    my ($a, $b) = @_;
    sqrt (($a->[0] - $b->[0])**2 + ($a->[1] - $b->[1])**2);
}

sub _angle_2d
{
    my ($a, $b) = @_;
    return ($PI/2) if (($a->[0] == $b->[0]) and ($a->[1] < $b->[1]));
    return ($PI/-2) if (($a->[0] == $b->[0]) and ($a->[1] > $b->[1]));
    my $radians = atan (($b->[1] - $a->[1]) / ($b->[0] - $a->[0]));
    $radians += $PI if $b->[0] < $a->[0];

    return $radians;
}

sub _subtract_2d
{
    my ($a, $b) = @_;
    return [$a->[0] - $b->[0], $a->[1] - $b->[1]];
}

sub _add_2d
{
    my ($a, $b) = @_;
    return [$b->[0] + $a->[0], $b->[1] + $a->[1]];
}

sub _scale_2d
{
    my ($factor, $a) = @_;
    ($factor, $a) = ($a, $factor) if ref $factor;
    return [$a->[0] * $factor, $a->[1] * $factor];
}

sub _normalise_2d
{
    my $a = shift;
    my $length = _distance_2d ([0,0], $a);
    $length = 0.00000000001 unless $length;
    return [$a->[0] / $length, $a->[1] / $length];
}

sub _det
{
    my ($a, $b) = @_;
    return ($a->[0] * $b->[1]) - ($a->[1] * $b->[0]);
}

# line math

sub _points_2line
{
    my ($coor_0, $coor_1) = @_;
    my $vector = _subtract_2d ($coor_1, $coor_0);
    $vector->[0] = 0.00000000001 unless $vector->[0];
    my $a = $vector->[1] / $vector->[0];
    my $b = $coor_0->[1] - ($coor_0->[0] * $a);
    {a => $a, b => $b};
}

sub _perpendicular_line
{
    my ($line, $coor) = @_;
    $line->{a} = 0.00000000001 unless $line->{a};
    my $a = -1/$line->{a};
    my $b = $coor->[1] - ($coor->[0] * $a);
    {a => $a, b => $b};
}

sub _line_intersection
{
    my ($line_0, $line_1) = @_;
    return undef if ($line_0->{a} == $line_1->{a});
    my $x = ($line_1->{b} - $line_0->{b}) / ($line_0->{a} - $line_1->{a});
    my $y = ($line_0->{a} * $x) + $line_0->{b};
    [$x, $y];
}

sub _perpendicular_distance
{
    my ($line, $coor) = @_;
    my $perpendicular_line = _perpendicular_line ($line, $coor);
    my $intersection = _line_intersection ($line, $perpendicular_line);
    _distance_2d ($coor, $intersection);
}

=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

=head1 SEE ALSO

L<Urb> L<File::DXF> L<molior.pl>

=head1 AUTHOR

Bruno Postle - 2011.

=cut

1;
